package ru.dbotthepony.kstarbound.util

import com.github.benmanes.caffeine.cache.Interner
import it.unimi.dsi.fastutil.HashCommon
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import ru.dbotthepony.kstarbound.stream
import java.lang.ref.ReferenceQueue
import java.lang.ref.WeakReference
import java.util.concurrent.locks.LockSupport
import java.util.stream.StreamSupport

class HashTableInterner<T : Any>(private val segmentBits: Int) : Interner<T> {
	companion object {
		private val interners = ArrayList<WeakReference<HashTableInterner<*>>>()

		private fun run() {
			var wait = 1_000_000L
			val minWait = 1_000_000L
			val maxWait = 1_000_000_000L

			while (true) {
				var any = 0

				synchronized(interners) {
					val i = interners.iterator()

					for (v in i) {
						val get = v.get()

						if (get == null) {
							i.remove()
						} else {
							for (segment in get.segments) {
								synchronized(segment) {
									any += segment.cleanup()
								}
							}
						}
					}
				}

				if (any != 0) {
					wait = (wait - 1_000_000L * any).coerceAtLeast(minWait)
				} else {
					wait = (wait + 1_000_000L).coerceAtMost(maxWait)
				}

				LockSupport.parkNanos(wait)
			}
		}

		private val thread = Thread(::run, "Interner Cleanup")

		init {
			thread.priority = 2
			thread.isDaemon = true
			thread.start()
		}
	}

	class Ref<T : Any>(referent: T, queue: ReferenceQueue<T>) : WeakReference<T>(referent, queue), Iterable<Ref<T>> {
		val hash = referent.hashCode()
		var nextEntry: Ref<T>? = null

		override fun hashCode(): Int {
			return hash
		}

		override fun equals(other: Any?): Boolean {
			return other === this
		}

		fun insert(value: Ref<T>) {
			var current = this

			while (current.nextEntry != null) {
				current = current.nextEntry!!
			}

			current.nextEntry = value
		}

		fun remove(value: Ref<T>): Boolean {
			var current = this

			while (true) {
				if (current.nextEntry === value) {
					current.nextEntry = current.nextEntry!!.nextEntry
					return true
				} else if (current.nextEntry == null) {
					return false
				} else {
					current = current.nextEntry!!
				}
			}
		}

		override fun iterator(): Iterator<Ref<T>> {
			return object : Iterator<Ref<T>> {
				private var next: Ref<T>? = this@Ref

				override fun hasNext(): Boolean {
					return next != null
				}

				override fun next(): Ref<T> {
					val next = next ?: throw NoSuchElementException()
					this.next = next.nextEntry
					return next
				}
			}
		}
	}

	private val actualSegmentBits: Int

	init {
		var result = 0

		for (i in 0 until segmentBits) {
			result = result or (1.shl(i))
		}

		actualSegmentBits = result
	}

	private val segments: Array<Segment> = Array(1.shl(segmentBits)) { Segment() }

	init {
		synchronized(interners) {
			interners.add(WeakReference(this))
		}
	}

	// chained linking hash table backed by plain array
	// while this increase memory usage (linked list), this greatly
	// simplify logic, and make scanning a bit faster because we don't jump to neighbour nodes
	// (assuming past our neighbour there is no such key)
	private inner class Segment : Interner<T> {
		private var mask = 31
		private var mem = arrayOfNulls<Ref<T>>(32)
		private var stored = 0
		private val queue = ReferenceQueue<T>()

		fun cleanup(): Int {
			var any = 0

			while (true) {
				val p = queue.poll() as Ref<T>? ?: return any
				remove(p)
				any++
			}
		}

		private fun hash(e: Any): Int {
			return HashCommon.mix(e.hashCode().rotateRight(segmentBits)) and mask
		}

		private fun grow() {
			mask = (mask shl 1) or 1

			val old = mem.stream()
				.filter { it != null }
				.flatMap { ObjectArrayList(it!!.iterator()).stream() }
				.filter { !it.refersTo(null) }
				.collect(ObjectArrayList.toList())

			for (elem in old) {
				elem.nextEntry = null
			}

			mem = arrayOfNulls(mem.size shl 1)
			val mem = mem

			for (elem in old) {
				val ehash = hash(elem)
				val existing = mem[ehash]

				if (existing == null) {
					mem[ehash] = elem
				} else {
					existing.insert(elem)
				}
			}
		}

		fun remove(ref: Ref<T>): Boolean {
			val hash = hash(ref)
			val mem = mem
			val search = mem[hash]

			if (search == null) {
				return false
			} else if (search === ref) {
				mem[hash] = search.nextEntry
				check(--stored >= 0) { "Stored count turned negative" }
				return true
			} else {
				return search.remove(ref).also {
					if (it) check(--stored >= 0) { "Stored count turned negative" }
				}
			}
		}

		override fun intern(sample: T): T {
			if (stored >= mem.size * 0.75f) grow()
			val mem = mem
			val hash = hash(sample)
			var search = mem[hash]

			while (search != null) {
				val get = search.get()

				if (get == sample)
					return get
				else if (get == null) {
					check(remove(search)) { "Unable to remove null entry $search at hash $hash" }
					search = search.nextEntry
				} else {
					search = search.nextEntry
				}
			}

			val ref = Ref(sample, queue)
			val existing = mem[hash]

			if (existing == null)
				mem[hash] = ref
			else
				existing.insert(ref)

			stored++
			return sample
		}
	}

	override fun intern(sample: T): T {
		val hash = HashCommon.mix(sample.hashCode())
		val segment = segments[hash and actualSegmentBits]

		synchronized(segment) {
			return segment.intern(sample)
		}
	}
}
